import logging

from gym import monitoring
from universe.vectorized import core  # Cannot import vectorized directly without inducing a cycle
from universe.wrappers.time_limit import TimeLimit

logger = logging.getLogger(__name__)

class _Monitor(core.Wrapper):
    def __init__(self, env, directory, video_callable=None, force=False,
                 resume=False, write_upon_reset=False, uid=None, mode=None):
        super(_Monitor, self).__init__(env)
        self.directory = directory
        self.video_callable = video_callable
        self.force = force
        self.resume = resume
        self.write_upon_reset = write_upon_reset
        self.uid = uid
        self.mode = mode

        # TODO if we want to monitor more than one instance in a vectorized
        # env we'll have to do this after configure()
        self._start_monitor()

    def _start_monitor(self):
        logger.info("Starting Monitor. Writing monitor logs to {}".format(self.directory))

        # Circular dependencies :(
        from universe import wrappers
        # We need to maintain pointers to these to avoid them being
        # GC'd. They have a weak reference to us to avoid cycles.
        # TODO if we want to monitor more than one instance in a vectorized
        # env we'll need to actually fix WeakUnvectorize
        self._unvectorized_envs = [wrappers.WeakUnvectorize(self, i) for i in range(1)]

        # For now we only monitor the first env
        self._monitor = monitoring.MonitorManager(self._unvectorized_envs[0])
        self._monitor.start(
            self.directory,
            self.video_callable,
            self.force,
            self.resume,
            self.write_upon_reset,
            self.uid,
            self.mode,
        )

    def _step(self, action_n):
        self._monitor._before_step(action_n[0])
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        done_n[0] = self._monitor._after_step(observation_n[0], reward_n[0], done_n[0], info)
        return observation_n, reward_n, done_n, info

    def _reset(self):
        self._monitor._before_reset()
        observation_n = self.env.reset()
        self._monitor._after_reset(observation_n[0])
        return observation_n

    def _close(self):
        super(_Monitor, self)._close()
        self._monitor.close()

    def set_monitor_mode(self, mode):
        logger.info("Setting the monitor mode is deprecated and will be removed soon")
        self._monitor._set_mode(mode)

def Monitor(env, directory, video_callable=None, force=False, resume=False,
            write_upon_reset=False, uid=None, mode=None):
    return _Monitor(TimeLimit(env), directory, video_callable, force, resume,
                    write_upon_reset, uid, mode)



import logging

import gym
from universe import error, spaces
from universe import vectorized

logger = logging.getLogger(__name__)


def atari_vnc(up=False, down=False, left=False, right=False, z=False):
    return [spaces.KeyEvent.by_name('up', down=up),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right),
            spaces.KeyEvent.by_name('down', down=down),
            spaces.KeyEvent.by_name('z', down=z)]

def slither_vnc(space=False, left=False, right=False):
    return [spaces.KeyEvent.by_name('space', down=space),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right)]

def racing_vnc(up=False, left=False, right=False):
    return [spaces.KeyEvent.by_name('up', down=up),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right)]

def platform_vnc(up=False, left=False, right=False, space=False):
    return [spaces.KeyEvent.by_name('up', down=up),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right),
            spaces.KeyEvent.by_name('space', down=space)]

def gym_core_action_space(gym_core_id):
    spec = gym.spec(gym_core_id)

    if spec.id == 'CartPole-v0':
        return spaces.Hardcoded([[spaces.KeyEvent.by_name('left', down=True)],
                                 [spaces.KeyEvent.by_name('left', down=False)]])
    elif spec._entry_point.startswith('gym.envs.atari:'):
        actions = []
        env = spec.make()
        for action in env.unwrapped.get_action_meanings():
            z = 'FIRE' in action
            left = 'LEFT' in action
            right = 'RIGHT' in action
            up = 'UP' in action
            down = 'DOWN' in action
            translated = atari_vnc(up=up, down=down, left=left, right=right, z=z)
            actions.append(translated)
        return spaces.Hardcoded(actions)
    else:
        raise error.Error('Unsupported env type: {}'.format(spec.id))



class SafeActionSpace(vectorized.Wrapper):
    """
    Recall that every universe environment receives a list of VNC events as action.
    There exist many environments for which the set of relevant action is much smaller
    and is known.   For example, Atari environments have a modest number of keys,
    so this wrapper, when applied to an Atari environment will reduce its action space.
    Doing so is very convenient for research, since today's RL algorithms rely on random
    exploration, which is hurt by small action spaces.  As our algorithms get better
    and we switch to using the raw VNC commands, this wrapper will become less important.


    NOTE: This class will soon be moved to `wrappers.experimental`. However the logic must currently remain in
    wrappers.SafeActionSpace in order to maintain backwards compatibility.
    """
    def __init__(self, env):
        super(SafeActionSpace, self).__init__(env)
        self._deprecation_warning()

        if self.spec.tags.get('runtime') == 'gym-core':
            self.action_space = gym_core_action_space(self.spec._kwargs['gym_core_id'])
        elif self.spec is None:
            pass
        elif self.spec.id in ['internet.SlitherIO-v0',
                              'internet.SlitherIOErmiyaEskandaryBot-v0',
                              'internet.SlitherIOEasy-v0']:
            self.action_space = spaces.Hardcoded([slither_vnc(left=True),
                                                  slither_vnc(right=True),
                                                  slither_vnc(space=True),
                                                  slither_vnc(left=True, space=True),
                                                  slither_vnc(right=True, space=True)])
        elif self.spec.id in ['flashgames.DuskDrive-v0']:
            # TODO: be more systematic
            self.action_space = spaces.Hardcoded([racing_vnc(up=True),
                                                  racing_vnc(left=True),
                                                  racing_vnc(right=True)])
        elif self.spec.id in ['flashgames.RedBeard-v0']:
            self.action_space = spaces.Hardcoded([platform_vnc(up=True),
                                                  platform_vnc(left=True),
                                                  platform_vnc(right=True),
                                                  platform_vnc(space=True)])

    def _deprecation_warning(self):
        logger.warn(('DEPRECATION WARNING: wrappers.SafeActionSpace has been moved to '
                     'wrappers.experimental.action_space.SafeActionSpace as of 2017-01-07. '
                     'Using legacy wrappers.SafeActionSpace will soon be removed'))



import logging

from universe.wrappers.action_space import SafeActionSpace as _SafeActionSpace

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class SafeActionSpace(_SafeActionSpace):
    """
    Recall that every universe environment receives a list of VNC events as action.
    There exist many environments for which the set of relevant action is much smaller
    and is known.   For example, Atari environments have a modest number of keys,
    so this wrapper, when applied to an Atari environment will reduce its action space.
    Doing so is very convenient for research, since today's RL algorithms rely on random
    exploration, which is hurt by small action spaces.  As our algorithms get better
    and we switch to using the raw VNC commands, this wrapper will become less important.

    NOTE: This will be the new location for SafeActionSpace, however the logic must currently remain in
    wrappers.SafeActionSpace in order to maintain backwards compatibility.
    """

    def _deprecation_warning(self):
        # No deprecation warning here because we are using the correct import
        pass



import logging

import time
from universe import pyprofile
from universe.vectorized import core

logger = logging.getLogger(__name__)

DEFAULT_MAX_EPISODE_SECONDS = 20 * 60.  # Default to 20 minutes if there is no explicit limit


class TimeLimit(core.Wrapper):
    def _configure(self, **kwargs):
        super(TimeLimit, self)._configure(**kwargs)
        self._max_episode_seconds = self.env.spec.tags.get('wrapper_config.TimeLimit.max_episode_seconds', None)
        self._max_episode_steps = self.env.spec.tags.get('wrapper_config.TimeLimit.max_episode_steps', None)

        if self._max_episode_seconds is None and self._max_episode_steps is None:
            self._max_episode_seconds = DEFAULT_MAX_EPISODE_SECONDS

        self._elapsed_steps = 0
        self._episode_started_at = None

    @property
    def _elapsed_seconds(self):
        return time.time() - self._episode_started_at

    def _past_limit(self):
        """Return true if we are past our limit"""
        if self._max_episode_steps is not None and self._max_episode_steps <= self._elapsed_steps:
            logger.debug("Env has passed the step limit defined by TimeLimit.")
            return True

        if self._max_episode_seconds is not None and self._max_episode_seconds <= self._elapsed_seconds:
            logger.debug("Env has passed the seconds limit defined by TimeLimit.")
            return True

        return False

    def _step(self, action_n):
        assert self._episode_started_at is not None, "Cannot call env.step() before calling reset()"
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        self._elapsed_steps += 1

        if self._past_limit():
            _ = self.reset()  # Force a reset, discard the observation
            done_n = [True] * self.n  # Force a done = True

        return observation_n, reward_n, done_n, info

    def _reset(self):
        self._episode_started_at = time.time()
        self._elapsed_steps = 0
        return self.env.reset()



import logging

from universe import vectorized, runtime_spec

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def CropObservations(env):
    """"
    Crops the visual observations of an environment so that they only contain the game screen.
    Removes anything outside the game that usually belongs to universe (browser borders and so on).
    """
    if env.spec.tags.get('flashgames', False):
        spec = runtime_spec('flashgames').server_registry[env.spec.id]
        return _CropObservations(env, x=18, y=84, height=spec["height"], width=spec["width"])
    elif (env.spec.tags.get('atari', False) and env.spec.tags.get('vnc', False)):
        return _CropObservations(env, height=194, width=160)
    else:
        # if unknown environment (or local atari), do nothing
        return env

class _CropObservations(vectorized.ObservationWrapper):
    def __init__(self, env, height, width, x=0, y=0):
        super(_CropObservations, self).__init__(env)
        self.x = x
        self.y = y
        self.height = height
        self.width = width

        # modify observation_space? (if so, how to know depth and channels before we have seen the first frame?)
        # self.observation_space = Box(0, 255, shape=(height, width, 3))

    def _observation(self, observation_n):
        return [self._crop_frame(observation) for observation in observation_n]

    def _crop_frame(self, frame):
        if frame is not None:
            if isinstance(frame, dict):
                frame['vision'] = frame['vision'][self.y:self.y + self.height, self.x:self.x + self.width]
            else:
                frame = frame[self.y:self.y + self.height, self.x:self.x + self.width]
        return frame
